package main

import (
        "math"
        "strconv"
        "fmt"
        "os"
)

type PopulationLattice struct {
    populationSize int
    lattice [][]*Agent
    individualCounter int
    globalLexicon *int
    agentLog map[int]string
}

func NewPopulationLattice() *PopulationLattice {
    globalLexicon := 0
    return &PopulationLattice{
        populationSize:         GlobalOptions.populationSize,
        globalLexicon:          &globalLexicon,
        agentLog:               make(map[int]string),
        individualCounter:      0}
}

func (pl *PopulationLattice) Populate() {
    latticeSize := int(math.Sqrt(float64(pl.populationSize)))
    Log("Populateing matrix, with each side " + strconv.Itoa(latticeSize))
    pl.lattice = make([][]*Agent, latticeSize)
    for x := range pl.lattice {
        pl.lattice[x] = make([]*Agent, latticeSize)
        for y := range pl.lattice[x] {
            agent := NewAgent(pl.individualCounter, x, y, true, true, pl.globalLexicon)
            pl.lattice[x][y] = agent
            pl.individualCounter+=1
        }
    }
}

func (pl *PopulationLattice) Size() int {
    return pl.populationSize
}

func (pl *PopulationLattice) Age() {
    for i := range pl.lattice {
        for j := range pl.lattice[i] {
            pl.lattice[i][j].age+=1
        }
    }
}



func (pl *PopulationLattice) GetGenerationStats() GenerationStats {
    agentList := pl.MakeListOfAgents()
    stats := NewStats()
    return stats.GenerateStatsFromGeneration(agentList)
}



func (pl *PopulationLattice) PrintRepresentation() {
    for i := range pl.lattice {
        for j := range pl.lattice[i] {
            if pl.lattice[i][j].alive == false {
                fmt.Printf("%v: %s ", pl.lattice[i][j].id, "Dead....")
            } else {
                fmt.Printf("%v: %f ", pl.lattice[i][j].id, pl.lattice[i][j].GetFitness())
            }
        }
        fmt.Print("\n")
    }
}

func (pl *PopulationLattice) Interact() {
    maxIterations := pl.Size() * GlobalOptions.dialogPrAgentPrGeneration
    for i := 0; i < maxIterations; i++ {
        agent1 := pl.GetRandomAgent()
        var agent2 *Agent
        // ensure not selection the same agent
        for true {
            //agent2 = pl.GetRandomAgent()
            neighbours := pl.GetNeighbours(agent1, true)
            agent2 = neighbours[randInt(0, len(neighbours)-1)]
            if agent1 != agent2 {
                break
            }
        }
        // Random agents speaks
        utterance := agent1.Speak();
        // Random neighbour hears
        success, learningCost := agent2.Hear(utterance)
        // update agent1s memory 
        // in lack of a better name
        agent1.DialogEnd(success, utterance, learningCost)
       // if(agent1.GetId() == 500) {
       //     fmt.Println(agent1)
       // }
       // if (randFloat() < 0.0000001) {
       //     fmt.Println("Random dialog, agent1 says " + string(utterance) + " and learning cost is " + FloatToString(learningCost))
       //     if(success) { 
       //         fmt.Println("Sukess!")
       //     }
       // }
    }
}

func (pl *PopulationLattice) KillWeak() {
    turnamentSlection := int(GlobalOptions.turnamentSlection*float64(pl.Size()))
    turnamentWinners := int(GlobalOptions.turnamentWinners*float64(pl.Size()))
    // put alle agents in a one dimention array for generic operations
    agents := pl.MakeListOfAgents()
    // shuffel the list
    agents = ShuffelAgentArray(agents)
    // get random k agents to turnament 
    turnamentAgents := agents[0:turnamentSlection]
    // sort them after fitness and choose the w best 
    turnamentAgents = SortAgentsByFitness(turnamentAgents)
    for i := 0; i < turnamentWinners; i++ {
        turnamentAgents[i].kill = false // dont kill the choosen ones
    }
    for a := range agents {
        if agents[a].kill == true {
            agents[a].alive = false // flagged as dead, 
        }
        agents[a].kill = true // set everyone as candidates for next generation killing
    }
    //pl.PrintRepresentation()
}

func (pl *PopulationLattice) Breed() {
    for x := range pl.lattice {
        for y := range pl.lattice[x] {
            if pl.lattice[x][y].alive == false {
                // we got a dead one! replace it!
                var agent *Agent
                // getting alive neighbours suiteable as parents 
                var parent1, parent2 *Agent
                aliveNeighbours := pl.GetNeighbours(pl.lattice[x][y], true)
                if len(aliveNeighbours) == 0 {
                    // if no parents nearby spawn random? cornercass, copy neighbour
                    agent = NewAgent(pl.individualCounter, x, y, true, true, pl.globalLexicon)
                    fmt.Println("New agent created in lack of neighbours at ", x, y)
                } else {
                    var genome *Genome
                    // find parents
                    if(len(aliveNeighbours) == 1) {
                        // only one parent
                        parent1 = aliveNeighbours[0] 
                        genome = parent1.genome.Copy()
                    } else {
                        parent1, parent2 = pl.AdultSelection(aliveNeighbours)
                        // copy parent1 and crossover with parent2
                        genome = parent1.genome.Copy()
                        genome.Crossover(parent2.genome)
                    }
                    // mutate with a given chance
                    if randFloat() < GlobalOptions.mutationChance {
                        genome.Mutate()
                    }
                    agent = NewAgent(pl.individualCounter, x, y, true, true, pl.globalLexicon)
                    agent.NewGenome(genome)
                    if(randFloat() > GlobalOptions.mutationChance) {
                        // if not mutation, inherited uttarnce
                        // else, invent new uttarnce in first dialog
                        inheritedUtterance := parent1.GetHighestWeightedUtterance()
                        //if (inheritedUtterance > 0) {
                            agent.GiveFirstUtterance(inheritedUtterance)
                        //}
                    }
                }
                pl.lattice[x][y] = agent
                pl.individualCounter+=1
            }
        }
    }
    //fmt.Println("---")
    //pl.PrintRepresentation()
}


func (pl *PopulationLattice) AdultSelection(neighbours []*Agent) (*Agent, *Agent) {
    neighbours = SortAgentsByFitness(neighbours)
    return neighbours[0], neighbours[1]
}

func (pl *PopulationLattice) StatsSnapshot() (*Snapshot) {
    latticeSize := int(math.Sqrt(float64(pl.populationSize)))
    latticeCopy := make([][]*Agent, latticeSize)
    for x := range latticeCopy {
        latticeCopy[x] = make([]*Agent, latticeSize)
        for y := range latticeCopy[x] {
            latticeCopy[x][y] = pl.lattice[x][y].Copy()
        }
    }
    snapshot := NewSnapshot("lattice")
    snapshot.AddLatticeData(latticeCopy)
    return snapshot
}

// help functions

func (pl *PopulationLattice) GetRandomAgent() (*Agent) {
    matrixSize := len(pl.lattice)
    count := 0
    for true {
        x, y := randInt(0,matrixSize-1), randInt(0,matrixSize-1)
        if pl.lattice[x][y].alive {
            return pl.lattice[x][y]
        }
        count+=1
        if count > 100 {
            msg := "Trying to locate alive agent but looks like everyone is dead. Exiting."
            fmt.Println(msg)
            Log(msg)
            os.Exit(1)
        }
    }
    return nil
}

func (pl *PopulationLattice) MakeListOfAgents() ([]*Agent) {
    counter := 0
    populationSize := len(pl.lattice)*len(pl.lattice)
    agents := make([]*Agent, populationSize)
    for i := range pl.lattice {
        for j := range pl.lattice[i] {
            agents[counter] = pl.lattice[i][j]
            counter+=1
        }
    }
    return agents
}

func (pl *PopulationLattice) GetNeighbours(agent *Agent, returnOnlyAlive bool) ([]*Agent) {
    var neighbours [9]*Agent
    matrixSize := len(pl.lattice[0])
    valid := 0
    for x := -1; x <= 1; x++ {
        for y := -1; y <= 1; y++ {
            if (x == 0 && y == 0)  || x+agent.x < 0 || x+agent.x >= matrixSize ||
                           y+agent.y < 0 || y+agent.y >= matrixSize {
                continue
            }
            if(returnOnlyAlive) {
                if !pl.lattice[agent.x+x][agent.y+y].alive {
                    continue
                }
            } else {
                if pl.lattice[agent.x+x][agent.y+y].alive {
                    continue
                }
            }
            neighbours[valid] = pl.lattice[agent.x+x][agent.y+y]
            valid++
        }
    }
    return neighbours[0:valid]
}


func (pl *PopulationLattice) Consensus() bool {
    utterance := pl.lattice[0][0].GetHighestWeightedUtterance()
    var newUtterance Utterance
    for i := range pl.lattice {
        for j := range pl.lattice[i] {
            newUtterance = pl.lattice[i][j].GetHighestWeightedUtterance()
            if utterance != newUtterance {
                return false
            }
        }
    }
    return true
}
